from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Tuple

from nys_constants.nys_constants import SortOrder
from nys_schemas.events_schema import (
    EventsResponse,
    EventsFilter,
    EventsSort,
)

from .base import BaseResource

__all__ = [
    'EventResource',  # The main resource class
    'EventsResponse',
    'EventsFilter',
    'EventsSort',
    'SortOrder',
]

class EventResource(BaseResource[EventsResponse]):
    """Resource for interacting with Events."""

    def search(
        self,
        *,
        search: Optional[str] = None,
        filters: Optional[Union[Dict[str, Any], EventsFilter]] = None,
        sort_by: Optional[Union[str, List[Tuple[EventsSort, SortOrder]]]] = None,
        page: int = 1,
        size: int = 50,
    ) -> List[EventsResponse]:
        """
        Search and filter Events.
        
        Args:
            search: search for keywords in the columns
                - Example: "component:brain OR description:error"
            filters: Event-specific filters. Can be either:
                - A EventsFilter instance with fields like hex_value, component, etc.
                - A dict matching the EventsFilter schema
            sort_by: a list of tuples of (EventsSort, SortOrder) or a string representing the sort order.
                - Example: [(EventsSort.CREATED_AT, SortOrder.DESC)]
                - Example: "-severity" (for descending order)
            page: Page number (1-based, default: 1)
            size: Items per page (default: 50)
            
        Returns:
            List[EventsResponse]: List of Events matching the filters
            
        Examples:
            >>> # Search using a dictionary filter
            >>> client.events.search(
            ...     filters={"component__eq": "brain", "severity__gte": 30}
            ... )
            
            >>> # Search using EventsFilter
            >>> from nys_client.events import EventsFilter, EventsSort, SortOrder
            >>> client.events.search(
            ...     filters=EventsFilter(component__eq="brain"),
            ...     sort_by=[(EventsSort.CREATED_AT, SortOrder.DESC)],
            ...     search="component:brain"
            ... )
            
            >>> # Search with string sort_by
            >>> client.events.search(
            ...     filters=EventsFilter(severity__gte=30),
            ...     sort_by="-created_at,severity",
            ...     search="component:brain"
            ... )
        """
        if isinstance(filters, dict):
            filters = EventsFilter(**filters)

        return self._search(
            endpoint="api/v2/events",
            response_model=EventsResponse,
            filters=filters,
            sort_by=sort_by,
            page=page,
            size=size,
            search=search
        )


from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Tuple

from nys_constants.nys_constants import OnEmptySKUAction, MeasurementUnit, SortOrder
from nys_schemas.inventory_view_schema import (
    InventoryViewResponse,
    InventoryViewFilter,
    InventoryViewSearch,
    InventoryViewSort,
)

from .base import BaseResource

__all__ = [
    'InventoryViewResource',  # The main resource class
    'InventoryViewResponse',
    'InventoryViewFilter',
    'InventoryViewSort',
    'InventoryViewSearch',
    'SortOrder',
    'OnEmptySKUAction',
    'MeasurementUnit',
]

class InventoryViewResource(BaseResource[InventoryViewResponse]):
    """Resource for interacting with Inventory."""

    def search(
        self,
        *,
        filters: Optional[Union[Dict[str, Any], InventoryViewFilter]] = None,
        search: Optional[str] = None,
        sort_by: Optional[Union[str, List[Tuple[InventoryViewSort, SortOrder]]]] = None,
        page: int = 1,
        size: int = 50
    ) -> List[InventoryViewResponse]:
        """
        Search and filter inventory.
        
        Args:
            filters: Inventory-specific filters. Can be either:
                - An InventoryViewFilter instance with fields like carrier_id, box_id, etc.
                - A dict matching the InventoryViewFilter schema
            search: Search for keywords in fields
                - Example: "carrier_id:123 AND box_id:456"
            sort_by: a list of tuples of (InventoryViewSort, SortOrder) or a string representing the sort order.
                - Example: [(InventoryViewSort.CARRIER_ID, SortOrder.DESC)]
                - Example: "-created_at" (for descending order)
            page: Page number (1-based, default: 1)
            size: Items per page (default: 50)
            
        Returns:
            List[InventoryViewResponse]: List of inventory matching the filters
            
        Examples:
            >>> # Search using InventoryViewFilter (recommended)
            >>> from nys_client.inventory_view import InventoryViewFilter, InventoryViewSort, InventoryViewSearch
            >>> client.inventory_view.search(
            ...     filters=InventoryViewFilter(
            ...         carrier_id__ilike="123",
            ...         box_id__ilike="456"
            ...     ),
            ...     search="carrier_id:123 AND box_id:456",
            ...     sort_by="-created_at",
            ... )

            >>> # Search using a dictionary filter
            >>> from nys_client.inventory_view import InventoryViewSort
            >>> client.inventory_view.search(
            ...     filters={"carrier_id": "123"},
            ...     search="carrier_id:123",
            ...     sort_by=[(InventoryViewSort.CARRIER_ID, SortOrder.DESC)],
            ... )
        """

        if isinstance(filters, dict):
            filters = InventoryViewFilter(**filters)

        if isinstance(sort_by, list):
            assert isinstance(sort_by[0][0], InventoryViewSort), "sort_by must be a list of Tuple[InventoryViewSort, SortOrder]"
            assert isinstance(sort_by[0][1], SortOrder), "sort_by must be a list of Tuple[InventoryViewSort, SortOrder]"
            assert len(sort_by[0]) == 2, "sort_by must be a list of Tuple[InventoryViewSort, SortOrder]"
        

        return self._search(
            endpoint="api/v2/inventory_view",
            response_model=InventoryViewResponse,
            filters=filters,
            search=search,
            sort_by=sort_by,
            page=page,
            size=size
        )

    def delete_load_by_box_id(self, box_id: int):
        """
        Delete a load by its box_id.
        
        Args:
            box_id: The ID of the load to delete
        """
        return self._client._request(
            method='DELETE',
            endpoint=f"api/v2/load/{box_id}",
            params={}
        )



from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Tuple
from uuid import UUID

from nys_constants.nys_constants import TriggerStatus, JobStatus, JobType, SortOrder
from nys_schemas.job_schema import (
    JobResponse,
    JobFilter,
    JobSort
)

from .base import BaseResource

__all__ = [
    'JobResource',  # The main resource class
    'JobResponse',
    'JobFilter',
    'JobSort',
    'SortOrder',
    'TriggerStatus',
    'JobStatus',
    'JobType',
]

class JobResource(BaseResource[JobResponse]):
    """Resource for interacting with jobs."""

    def search(
        self,
        *,
        filters: Optional[Union[Dict[str, Any], JobFilter]] = None,
        sort_by: Optional[Union[str, List[Tuple[JobSort, SortOrder]]]] = None,
        page: int = 1,
        size: int = 50
    ) -> List[JobResponse]:
        """
        Search and filter jobs.
        
        Args:
            filters: Job-specific filters. Can be either:
                - A JobFilter instance with fields like status__eq, type__eq, etc.
                - A dict matching the JobFilter schema
            sort_by: a list of tuples of (JobSort, SortOrder) or a string representing the sort order.
                - Example: [(JobSort.CREATED_AT, SortOrder.DESC)]
                - Example: "-created_at" (for descending order)
            page: Page number (1-based, default: 1)
            size: Items per page (default: 50)
            
        Returns:
            List[JobResponse]: List of jobs matching the filters
            
        Examples:
            Search using enums (recommended):
            >>> from nys_client.job import JobType, JobStatus, JobSort
            >>> client.jobs.search(
            ...     filters=JobFilter(status__eq=JobStatus.EXECUTING, type__eq=JobType.PICKING),
            ...     sort_by=[(JobSort.CREATED_AT, SortOrder.DESC)]
            ... )

            Search for executing picking jobs using string literals:
            >>> client.jobs.search(
            ...     filters={"status__eq": "EXECUTING", "type__eq": "PICKING"},
            ...     sort_by="-created_at",
            ... )
        """
        if isinstance(filters, dict):
            filters = JobFilter(**filters)

        return self._search(
            endpoint="api/v2/jobs",
            response_model=JobResponse,
            filters=filters,
            sort_by=sort_by,
            page=page,
            size=size
        )

    def trigger(
        self, 
        job_id: Union[str, UUID], 
        trigger_status: TriggerStatus = TriggerStatus.SUCCEEDED_TRIGGER
    ) -> None:
        """
        Trigger a specific job with a status update.
        
        Args:
            job_id: The ID of the job to trigger (string or UUID)
            trigger_status: The trigger status to set (default: SUCCEEDED_TRIGGER)
            
        Raises:
            requests.HTTPError: If the request fails
            
        Example:
            >>> # Trigger a job as succeeded
            >>> nys_cli.jobs.trigger(
            ...     job_id="job-123",
            ...     trigger_status=TriggerStatus.SUCCEEDED_TRIGGER
            ... )
        """
        payload = {
            "trigger_status": trigger_status
        }
        self._client._request(
            method="POST",
            endpoint=f"api/v2/trigger/{job_id}",
            json=payload
        ) 


from __future__ import annotations

from nys_schemas.load_schema import (
    LoadResponse,
    LoadCreate,
    LoadPatch,
)

from .base import BaseResource

__all__ = [
    'LoadResource',  # The main resource class
    'LoadResponse',
    'LoadCreate',
    'LoadPatch',
]

class LoadResource(BaseResource[LoadResponse]):
    """Resource for interacting with loads."""

    def create(
        self,
        box_id: int,
        load: LoadCreate
    ) -> LoadResponse:
        """
        Create a new load for a specific box.
        
        Args:
            box_id: ID of the box to create the load for
            load: LoadCreate object containing the load parameters
            
        Returns:
            LoadResponse: The created load object
            
        Example:
            >>> nys_cli.loads.create(
            ...     box_id=123,
            ...     load=LoadCreate(
            ...         sku_id="SKU123",
            ...         item_count=10,
            ...         item_count_max=20
            ...     )
            ... )
        """
        response = self._client._request(
            method="POST",
            endpoint=f"api/v2/load/{box_id}",
            json=load.dict(exclude_unset=True)
        )
        
        return LoadResponse(**response.json())

    def patch(
        self,
        box_id: int,
        patch: LoadPatch,
    ) -> LoadResponse:
        """
        Update an existing load for a specific box.
        Only fields that are explicitly provided in the patch will be updated.
        
        Args:
            box_id: ID of the box whose load to update
            patch: LoadPatch object containing the fields to update
            
        Returns:
            LoadResponse: The updated load object
            
        Example:
            >>> nys_cli.loads.patch(
            ...     box_id=123,
            ...     patch=LoadPatch(
            ...         sku_id="SKU456",
            ...         item_count=15,
            ...         item_count_max=25
            ...     )
            ... )
        """
        response = self._client._request(
            method="PATCH",
            endpoint=f"api/v2/load/{box_id}",
            json=patch.dict(exclude_unset=True)
        )
        
        return LoadResponse(**response.json())

    def delete(
        self,
        box_id: int,
    ) -> None:
        """
        Delete a load for a specific box.
        
        Args:
            box_id: ID of the box whose load to delete
            
        Example:
            >>> nys_cli.loads.delete(box_id=123)
        """
        response =self._client._request(
            method="DELETE",
            endpoint=f"api/v2/load/{box_id}"
        ) 
        return response


from __future__ import annotations
from typing import Any, List
from nys_schemas.system_schema import (
    StorageStatusResponse,
    StorageResponse,
    EnvFileResponse,
    ReportResponse,
    BalconyResponse,
    LevelResponse
)
from .base import BaseResource

class SystemResource(BaseResource[Any]):
    """Resource for interacting with storage-level endpoints."""

    def get_status(self) -> StorageStatusResponse:
        """
        Get the current storage system status.
        Returns:
            The storage status as a StorageStatusResponse object.
        """
        response = self._client._request(
            method="GET",
            endpoint="api/v2/storage_status"
        )
        # The API returns a dict with a 'storage_status' key
        return StorageStatusResponse(**response.json())

    def start(self) -> StorageResponse:
        """
        Start the brain system by calling the /v2/start_brain endpoint.
        Returns the API response as a StorageResponse object.
        """
        response = self._client._request(
            method="POST",
            endpoint="api/v2/start_brain",
        )
        return StorageResponse(**response.json())

    def stop(self) -> StorageResponse:
        """
        Stop the brain system by calling the /v2/stop_brain endpoint.
        Returns the API response as a StorageResponse object.
        """
        response = self._client._request(
            method="POST",
            endpoint="api/v2/stop_brain",
        )
        return StorageResponse(**response.json())
    
    def get_env_file(self) -> EnvFileResponse:
        """
        Get the environment file by calling the /v2/get_env_file endpoint.
        Returns the API response as an EnvFileResponse object.
        """
        response = self._client._request(
            method="GET",
            endpoint="api/v2/get_env_file",
        )
        return EnvFileResponse(**response.json())
    
    def get_report(self) -> ReportResponse:
        """
        Get the report by calling the /v2/get_report endpoint.
        Returns the API response as a ReportResponse object.
        """
        response = self._client._request(
            method="GET",
            endpoint="api/v2/report",
        )
        return ReportResponse(**response.json())
    
    def get_balconies(self) -> List[BalconyResponse]:
        """
        Get the list of balconies by calling the /v2/balconies endpoint.
        Returns the API response as a list of BalconyResponse objects.
        """
        response = self._client._request(
            method="GET",
            endpoint="api/v2/balconies",
        )
        return [BalconyResponse(**balcony) for balcony in response.json()]
    
    def get_levels(self) -> List[LevelResponse]:
        """
        Get the list of levels by calling the /v2/levels endpoint.
        Returns the API response as a list of LevelResponse objects.
        """
        response = self._client._request(
            method="GET",
            endpoint="api/v2/levels",
        )
        return [LevelResponse(**level) for level in response.json()]
    
    def resume_requests(self) -> StorageResponse:
        """
        TODO deprecate
        Resume requests by calling the /v2/resume_requests endpoint.
        Returns the API response as a StorageResponse object.
        """
        response = self._client._request(
            method="POST",
            endpoint="api/v2/resume_requests",
        )
        return StorageResponse(**response.json())
    


from __future__ import annotations
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union, Tuple
from pydantic import BaseModel

from nys_constants.nys_constants import SortOrder
from .utils import prepare_filter_value

T = TypeVar('T', bound=BaseModel)
FilterType = TypeVar('FilterType', bound=BaseModel)

class BaseResource(Generic[T]):
    """Base class for API resources. e.g. Job, Request, Sku, etc"""
    
    def __init__(self, client: 'NysClient'):
        self._client = client

    #TODO consider renaming to _get
    def _search(
        self,
        *,
        endpoint: str,
        response_model: Type[T],
        filters: Optional[FilterType] = None,
        search: Optional[str] = None,
        sort_by: Optional[Union[str, List[Tuple[str, SortOrder]]]] = None,
        page: int = 1,
        size: int = 50
    ) -> List[T]:
        """Generic search method used by all resource-specific search methods."""
        params: Dict[str, Any] = {
            "page": page,
            "size": size
        }

        # Add sorting parameters if provided
        if sort_by:
            if isinstance(sort_by, str):
                params["sort_by"] = sort_by
            else:
                # Convert list of tuples to comma-separated string with prefixes
                sort_str = ','.join(
                    f"{'-' if order == SortOrder.DESC else ''}{field}"
                    for field, order in sort_by
                )
                params["sort_by"] = sort_str

        # Add search fields if provided
        if search:
            if isinstance(search, str):
                # leave the validation to the API
                params['search'] = search
            else:
                raise ValueError("search must be a string")

        # Add filters if provided
        if filters:
            # Convert filter model to dict and remove None values
            filter_dict = filters.dict(exclude_unset=True, exclude_defaults=True)
            
            # Add each non-None filter to params, converting values as needed
            for field_name, value in filter_dict.items():
                if value is not None:
                    params[field_name] = prepare_filter_value(value)

        # Make API request
        response = self._client._request(
            method="GET",
            endpoint=endpoint,
            params=params
        )
        
        # Parse response into list of model instances
        data = response.json()
        return [response_model(**item) for item in data["items"]] 


from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Tuple

from nys_constants.nys_constants import OnEmptySKUAction, MeasurementUnit, SortOrder
from nys_schemas.sku_schema import (
    Sku,
    SkuCreate,
    SkuResponse,
    SkuFilter,
    SkuPatch,
    SkuSort,
)

from .base import BaseResource

__all__ = [
    'SkuResource',  # The main resource class
    'Sku',
    'SkuCreate',
    'SkuResponse',
    'SkuFilter',
    'SkuPatch',
    'SkuSort',
    'SortOrder',
    'OnEmptySKUAction',
    'MeasurementUnit',
]

class SkuResource(BaseResource[SkuResponse]):
    """Resource for interacting with SKUs."""

    def create(
        self,
        sku: SkuCreate
    ) -> SkuResponse:
        """
        Create a new SKU with the specified parameters.
        
        Args:
            sku: SkuCreate object containing the SKU parameters
            
        Returns:
            Sku: The created SKU object
            
        Example:
            >>> from nys_client.skus import MeasurementUnit
            >>> client.skus.create(
            ...     sku=SkuCreate(
            ...         id="SKU123",
            ...         name="Test SKU",
            ...         description="A test SKU",
            ...         measurement_unit=MeasurementUnit.PIECES
            ...     )
            ... )
        """

        
        response = self._client._request(
            method="POST",
            endpoint="api/v2/sku",
            json=sku.dict(exclude_unset=True)
        )
        
        return SkuResponse(**response.json())

    def search(
        self,
        *,
        filters: Optional[Union[Dict[str, Any], SkuFilter]] = None,
        search: Optional[str] = None,
        sort_by: Optional[Union[str, List[Tuple[SkuSort, SortOrder]]]] = None,
        page: int = 1,
        size: int = 50
    ) -> List[SkuResponse]:
        """
        Search and filter SKUs.
        
        Args:
            filters: SKU-specific filters. Can be either:
                - A SkuFilter instance with fields like id__eq, name__ilike, etc.
                - A dict matching the SkuFilter schema
            search: Search query string to filter SKUs.
                - Field-specific search: field:value (e.g., id:tee)
                - Logical operators: AND, OR
                - Grouping with parentheses
                - Results will be ordered by the greatest trigram similarity among all the specified field:value pairs
                - Example: "id:tee OR name:Tee"
            sort_by: a list of tuples of (SkuSort, SortOrder) or a string representing the sort order.
                - Example: [(SkuSort.TOTAL_QUANTITY, SortOrder.DESC)]
                - Example: "-created_at" (for descending order)
            page: Page number (1-based, default: 1)
            size: Items per page (default: 50)
            
        Returns:
            List[SkuResponse]: List of SKUs matching the filters
            
        Examples:
            >>> # Search using a dictionary filter
            >>> from nys_client.skus import MeasurementUnit, SkuSort
            >>> client.skus.search(
            ...     filters={"measurement_unit__in": [MeasurementUnit.PIECE, MeasurementUnit.GRAM]},
            ...     search="id:Tee OR name:Tee",
            ...     sort_by=[(SkuSort.CREATED_AT, SortOrder.DESC)]
            ... )

            >>> # Search using SkuFilter (recommended)
            >>> from nys_client.skus import SkuFilter, MeasurementUnit, SkuSort
            >>> client.skus.search(
            ...     filters=SkuFilter(
            ...         measurement_unit__in=[MeasurementUnit.PIECE, MeasurementUnit.GRAM],
            ...         name__ilike="test"
            ...     ),
            ...     sort_by=[(SkuSort.CREATED_AT, SortOrder.DESC)]
            ... )
        """
        if isinstance(filters, dict):
            filters = SkuFilter(**filters)

        return self._search(
            endpoint="api/v2/skus",
            response_model=SkuResponse,
            filters=filters,
            search=search,
            sort_by=sort_by,
            page=page,
            size=size
        )

    def patch(
        self,
        sku_id: str,
        patch: SkuPatch,
    ) -> SkuResponse:
        """
        Update an existing SKU with the specified parameters.
        Only fields that are explicitly provided in the patch will be updated.
        
        Args:
            sku_id: ID of the SKU to update
            patch: SkuPatch object containing the fields to update
            
        Returns:
            Sku: The updated SKU object
            
        Example:
            >>> from nys_client.skus import SkuPatch, MeasurementUnit
            >>> client.skus.patch(
            ...     sku_id="SKU123",
            ...     patch=SkuPatch(
            ...         name="Updated SKU Name",
            ...         measurement_unit=MeasurementUnit.PIECE
            ...     )
            ... )
        """
        response = self._client._request(
            method="PATCH",
            endpoint=f"api/v2/sku/{sku_id}",
            json=patch.dict(exclude_unset=True)
        )
        
        return SkuResponse(**response.json()) 


from __future__ import annotations
from datetime import datetime
from enum import Enum
from typing import Any, List

def prepare_filter_value(value: Any) -> Any:
    """Prepare a filter value for API request."""
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, (list, tuple)):
        return [prepare_filter_value(v) for v in value]
    return value



from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Tuple
from datetime import datetime

from nys_constants.nys_constants import RequestType, RequestStatus, JobType, JobStatus, SortOrder
from nys_schemas.requests_and_jobs_view_schema import (
    RequestsAndJobsViewResponse,
    RequestsAndJobsViewFilter,
    RequestsAndJobsViewSort
)

from .base import BaseResource

__all__ = [
    'RequestsAndJobsViewResource',  # The main resource class
    'RequestsAndJobsViewResponse',
    'RequestsAndJobsViewFilter',
    'RequestsAndJobsViewSort',
    'SortOrder',
    'RequestType',
    'RequestStatus',
    'JobType',
    'JobStatus',
]

class RequestsAndJobsViewResource(BaseResource[RequestsAndJobsViewResponse]):
    """Resource for interacting with requests and jobs view."""

    def search(
        self,
        *,
        search: Optional[str] = None,
        filters: Optional[Union[Dict[str, Any], RequestsAndJobsViewFilter]] = None,
        sort_by: Optional[Union[str, List[Tuple[RequestsAndJobsViewSort, SortOrder]]]] = None,
        page: int = 1,
        size: int = 50
    ) -> List[RequestsAndJobsViewResponse]:
        """
        Search and filter requests and jobs from the materialized view.
        
        Args:
            search: search for keywords in the columns
                - Example: "type:PICKING AND status:EXECUTING"
            filters: Request and job specific filters. Can be either:
                - A RequestAndJobsViewFilter instance with fields like status, type, etc.
                - A dict matching the RequestAndJobsViewFilter schema
            sort_by: Field to sort by (from RequestAndJobsViewSort enum)
            page: Page number (1-based, default: 1)
            size: Items per page (default: 50)
            
        Returns:
            List[RequestAndJobsViewResponse]: List of requests and jobs matching the filters
            
        Examples:
            >>> # Using RequestAndJobsViewFilter (recommended)
            >>> from nys_client.requests_and_jobs_view import (
            ...     RequestAndJobsViewFilter,
            ...     RequestAndJobsViewSort,
            ...     JobType,
            ...     JobStatus,
            ...     RequestType,
            ...     RequestStatus
            ... )
            >>> nys_cli.requests_and_jobs_view.search(
            ...     filters=RequestAndJobsViewFilter(
            ...         job_status__eq=JobStatus.EXECUTING,
            ...         job_type__eq=JobType.PICKING,
            ...         request_type__eq=RequestType.FULFILLMENT,
            ...         request_status__eq=RequestStatus.EXECUTING
            ...     ),
            ...     sort_by=[(RequestAndJobsViewSort.CREATED_AT, SortOrder.DESC)]
            ... )

            >>> # Using dictionary with enums
            >>> nys_cli.requests_and_jobs_view.search(
            ...     filters={
            ...         "job_status__in": [JobStatus.EXECUTING, JobStatus.SUCCEEDED],
            ...         "job_type__eq": JobType.PICKING,
            ...         "request_type__eq": RequestType.FULFILLMENT
            ...     },
            ...     sort_by=[(RequestAndJobsViewSort.UPDATED_AT, SortOrder.DESC)]
            ... )
        """
        if isinstance(filters, dict):
            filters = RequestsAndJobsViewFilter(**filters)

        return self._search(
            endpoint="api/v2/requests_and_jobs",
            response_model=RequestsAndJobsViewResponse,
            filters=filters,
            sort_by=sort_by,
            page=page,
            size=size,
            search=search
        ) 


from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Tuple, Literal

from nys_constants.nys_constants import RequestType, RequestStatus, SortOrder
from nys_schemas.request_schema import (
    RequestFilter, 
    RequestSort,
    OnboardingRequestCreateInput,
    OffboardingRequestCreateInput,
    RFIDWriteRequestCreateInput,
    PauseRequestCreateInput,
    ResumeRequestCreateInput,
    FetchRequestCreateInput,
    FulfillmentRequestCreateInput,
    ReplenishmentRequestCreateInput,
    SkuEntityInput,
    CarrierEntityInput,
    RequestResponseComposite,
)

from .base import BaseResource

__all__ = [
    'RequestResource',  # The main resource class
    'RequestFilter',
    'RequestResponseComposite',
    'RequestSort',
    'RequestStatus',
    'RequestType',
    'FetchRequestCreateInput',
    'FulfillmentRequestCreateInput',
    'ReplenishmentRequestCreateInput',
]

class RequestResource(BaseResource[RequestResponseComposite]):
    """Resource for interacting with requests."""

    def create_fetch(
        self,
        entities: List[str],
        priority: Optional[int] = None,
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        request_input = FetchRequestCreateInput(
                type=RequestType.FETCH,
                priority=priority,
                id=id,
                entities=[CarrierEntityInput(carrier_id=carrier_id) for carrier_id in entities]
            )
        response = self._client._request(
            method="POST",
            endpoint=f"api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )
        return RequestResponseComposite(**response.json())
    
    def create_fulfillment(
        self,
        entities: List[Tuple[str, int]],
        priority: Optional[int] = None,
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        _entities = [SkuEntityInput(sku_id=sku_id, quantity=quantity) for sku_id, quantity in entities]
        rsp = self._create_fulfillment_or_replenishment_request(RequestType.FULFILLMENT, _entities, priority, id)
        return RequestResponseComposite(**rsp)
    
    def create_replenishment(
        self,
        entities: List[Tuple[str, int]],
        priority: Optional[int] = None,
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        _entities = [SkuEntityInput(sku_id=sku_id, quantity=quantity) for sku_id, quantity in entities]
        rsp = self._create_fulfillment_or_replenishment_request(RequestType.REPLENISHMENT, _entities, priority, id)
        return RequestResponseComposite(**rsp)
    
    def _create_fulfillment_or_replenishment_request(
        self,
        request_type: Literal[RequestType.FULFILLMENT, RequestType.REPLENISHMENT],
        entities: List[SkuEntityInput],
        priority: Optional[int] = None,
        id: Optional[str] = None,
    ) -> Any:
        assert request_type in [RequestType.FULFILLMENT, RequestType.REPLENISHMENT], f"Invalid request type: {request_type}"

        # Create the request input model
        if request_type == RequestType.FULFILLMENT:
            request_input = FulfillmentRequestCreateInput(
                type=request_type,
                priority=priority,
                id=id,
                entities=entities
            )
        elif request_type == RequestType.REPLENISHMENT:
            request_input = ReplenishmentRequestCreateInput(
                type=request_type,
                priority=priority,
                id=id,
                entities=entities
            )
        else:
            raise ValueError(f"Invalid request type: {request_type}")
        
        # Make the API request
        response = self._client._request(
            method="POST",
            endpoint=f"api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )

        return response.json()

    def search(
        self,
        *,
        filters: Optional[Union[Dict[str, Any], RequestFilter]] = None,
        sort_by: Optional[Union[List[Tuple[RequestSort, SortOrder]], str]] = None,
        page: int = 1,
        size: int = 50
    ) -> List[RequestResponseComposite]:
        """
        Search and filter requests.
        
        Args:
            filters: Request-specific filters. Can be either:
                - A RequestFilter instance with fields like status, type, etc.
                - A dict matching the RequestFilter schema
            sort_by: a list of tuples of (RequestSort, SortOrder) or a string representing the sort order.
                - Example: [(RequestSort.CREATED_AT, SortOrder.DESC)]
                - Example: "-created_at" (for descending order)
            page: Page number (1-based, default: 1)
            size: Items per page (default: 50)
            
        Returns:
            List[RequestResponse]: List of requests matching the filters
            
        Examples:
            Search for executing fulfillment requests using string literals:
            >>> nys_cli.requests.search(
            ...     filters={"status": "EXECUTING", "type": "FULFILLMENT"},
            ...     sort_by="-created_at,
            ... )

            Search using enums (recommended):
            >>> from nys_client.request import RequestType, RequestStatus
            >>> nys_cli.requests.search(
            ...     filters=RequestFilter(status__eq=RequestStatus.EXECUTING, type__eq=RequestType.FULFILLMENT),
            ...     sort_by=[(RequestSort.CREATED_AT, SortOrder.DESC)],
            ... )
        """
        if isinstance(filters, dict):
            filters = RequestFilter(**filters)
            

        return self._search(
            endpoint="api/v2/requests",
            response_model=RequestResponseComposite,
            filters=filters,
            sort_by=sort_by,
            page=page,
            size=size
        )

    def create_onboarding(
        self,
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        """
        Create a new onboarding request.
        
        Args:
            id: Optional ID (auto-generated if not provided)
            
        Returns:
            RequestResponse containing the created request
            
        Examples:
            >>> nys_cli.requests.create_onboarding()
        """ 
        # Create the request input model
        request_input = OnboardingRequestCreateInput(
            type=RequestType.ONBOARD,
            id=id,
        )
        
        # Make the API request
        response = self._client._request(
            method="POST",
            endpoint="api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )
        # Parse and return the response
        return RequestResponseComposite(**response.json())

    def create_offboarding(
        self,
        bot_id: int,
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        """
        Create a new offboarding request.
        
        Args:
            bot_id: ID of the bot to offboard
            id: Optional external ID (auto-generated if not provided)
            
        Returns:
            RequestResponse containing the created request
            
        Examples:
            >>> nys_cli.requests.create_offboarding(bot_id=1)
        """
        # Create the request input model
        request_input = OffboardingRequestCreateInput(
            type=RequestType.OFFBOARD,
            id=id,
            bot_id=bot_id,
        )
        
        # Make the API request
        response = self._client._request(
            method="POST",
            endpoint="api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )

        # Parse and return the response
        return RequestResponseComposite(**response.json())

    def create_rfid_write(
        self,
        level_id: int,
        action_type: str,
        start_from_scratch: Optional[bool],
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        """
        Create a new RFID write request.
        
        Args:
            level_id: ID of the level to write RFID tags on
            action_type: Type of RFID action (e.g., "write", "check", "write_and_check")
            start_from_scratch: Whether to start from scratch (default: True)
            id: Optional ID (auto-generated if not provided)
            
        Returns:
            RequestResponse containing the created request
            
        Examples:
            >>> nys_cli.requests.create_rfid_write(level_id=1, action_type="write")
            >>> nys_cli.requests.create_rfid_write(level_id=2, action_type="check", start_from_scratch=False)
        """
        # Create the request input model
        request_input = RFIDWriteRequestCreateInput(
            type=RequestType.RFID_MAINTENANCE,
            id=id,
            level_id=level_id,
            action_type=action_type,
            start_from_scratch=start_from_scratch
        )
        
        # Make the API request
        response = self._client._request(
            method="POST",
            endpoint="api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )

        # Parse and return the response
        return RequestResponseComposite(**response.json())

    def create_pause(
        self,
        id: Optional[str] = None,
    ) -> RequestResponseComposite:
        """
        Create a new level pause request.
        
        Args:
            id: Optional ID (auto-generated if not provided)
            
        Returns:
            RequestResponse containing the created request
            
        Examples:
            >>> nys_cli.requests.create_pause()
        """
        # Create the request input model
        request_input = PauseRequestCreateInput(
            type=RequestType.PAUSE,
            id=id,
        )
        
        # Make the API request
        response = self._client._request(
            method="POST",
            endpoint="api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )

        # Parse and return the response
        return RequestResponseComposite(**response.json())

    def create_resume(
        self,
        id: Optional[str] =None,
    ) -> RequestResponseComposite:
        """
        Create a new level resume request.
        
        Args:
            priority: Request priority (default: 100)
            id: Optional ID (auto-generated if not provided)
            
        Returns:
            RequestResponse containing the created request
            
        Examples:
            >>> nys_cli.requests.create_resume()
        """
        # Create the request input model
        request_input = ResumeRequestCreateInput(
            type=RequestType.RESUME,
            id=id,
        )
        
        # Make the API request
        response = self._client._request(
            method="POST",
            endpoint="api/v2/requests",
            json=request_input.dict(exclude_unset=True, exclude_none=True)
        )

        # Parse and return the response
        return RequestResponseComposite(**response.json())
    


from __future__ import annotations
import logging
import requests
from typing import Optional

from .request import RequestResource
from .job import JobResource
from .skus import SkuResource
from .loads import LoadResource
from .requests_and_jobs_view import RequestsAndJobsViewResource
from .inventory_view import InventoryViewResource
from .system import SystemResource
from .bot import BotResource
from .events import EventResource
logger = logging.getLogger(__name__)
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

class NysClient:
    """Client for interacting with the Noyes API."""
    
    def __init__(
        self,
        api_key: str,
        base_url: str = "http://noyes-api:8003",
        # TODO: Increased timeout to 60 seconds for long-running requests. Eg. stop_brain endpoint
        timeout: int = 60 # seconds
    ):
        """
        Initialize the NysClient.
        
        Args:
            api_key: Your API key
            base_url: Base URL of the API (default: http://noyes-api:8003)
            timeout: Request timeout in seconds (default: 60)
        """
        self.api_key = api_key
        self.base_url = base_url.rstrip("/")
        self.timeout = timeout

        self._headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }
        
        # Initialize resources
        self._requests = RequestResource(self)
        self._jobs = JobResource(self)
        self._skus = SkuResource(self)
        self._loads = LoadResource(self)
        self._requests_and_jobs_view = RequestsAndJobsViewResource(self)
        self._inventory_view = InventoryViewResource(self)
        self._system = SystemResource(self)
        self._bots = BotResource(self)
        self._events = EventResource(self)
    
    @property
    def requests(self) -> RequestResource:
        """Access request-related operations."""
        return self._requests
    
    @property
    def jobs(self) -> JobResource:
        """Access job-related operations."""
        return self._jobs

    @property
    def skus(self) -> SkuResource:
        """Access SKU-related operations."""
        return self._skus

    @property
    def loads(self) -> LoadResource:
        """Access load-related operations."""
        return self._loads

    @property
    def requests_and_jobs_view(self) -> RequestsAndJobsViewResource:
        """Access requests and jobs view operations."""
        return self._requests_and_jobs_view

    @property
    def inventory_view(self) -> InventoryViewResource:
        """Access inventory view operations."""
        return self._inventory_view

    @property
    def system(self) -> SystemResource:
        """Access system-related operations."""
        return self._system
        
    @property
    def bots(self) -> BotResource:
        """Access bot-related operations."""
        return self._bots

    @property
    def events(self) -> EventResource:
        """Access event-related operations."""
        return self._events

    def _request(
        self,
        method: str,
        endpoint: str,
        params: Optional[dict] = None,
        json: Optional[dict] = None,
        **kwargs
    ) -> requests.Response:
        """
        Make a request to the API.
        
        Args:
            method: HTTP method (GET, POST, etc.)
            endpoint: API endpoint
            params: Query parameters
            json: JSON body
            **kwargs: Additional arguments to pass to requests
            
        Returns:
            requests.Response: Response from the API
        """
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        response = requests.request(
            method=method,
            url=url,
            params=params,
            json=json,
            timeout=self.timeout,
            headers=self._headers,
            **kwargs
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.error(f"HTTP error occurred: {e}")
            logger.error(f"Response content: {response.text}")
            raise e
        return response


import os
from setuptools import setup, find_packages

setup(
    name="nys_client",
    version="0.1.1",
    packages=find_packages(),
    install_requires=[
        "requests>=2.28.1",
        "python-dotenv>=1.0.0",
        "typing-extensions>=4.12.2",
        "pydantic>=1.10.0,<2.0.0",
        "fastapi-pagination>=0.12.0",
        "nys_constants>=0.1.0",
        "nys_schemas>=0.1.0",
    ],
    python_requires=">=3.8",
    author="Noyes",
    author_email="support@noyes-tech.com",
    description="Python client library for the Noyes API",
    long_description=open("README.md").read() if os.path.exists("README.md") else "",
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
) 


from __future__ import annotations
from typing import Any, Dict, List, Optional
from nys_schemas.bot_schema import (
    Bot,
    BotResponse,
    BotFilter,
    BotSort,
    BotChargingRequest,
    BotChargingResponse,
    BotStopChargingRequest,
    BotStopChargingResponse,
    BotRemoveResponse
)
from .base import BaseResource

__all__ = [
    'BotResource',  # The main resource class
    'Bot',
    'BotResponse',
    'BotFilter',
    'BotSort',
    'BotChargingRequest',
    'BotChargingResponse',
    'BotStopChargingRequest',
    'BotStopChargingResponse',
    'BotRemoveResponse'
]

class BotResource(BaseResource[BotResponse]):
    """Resource for interacting with Bots."""

    def get_bots(self) -> List[BotResponse]:
        """
        Get all bots.
        
        Returns:
            List[BotResponse]: List of all bots in the system
            
        Examples:
            >>> bots = client.bots.get_bots()
            >>> for bot in bots:
            ...     print(f"Bot ID: {bot.idx}, Level: {bot.level_id}, Battery: {bot.battery_level}%")
        """
        response = self._client._request(
            method="GET",
            endpoint="api/v2/bots"
        )
        return [BotResponse(**item) for item in response.json()]

    def go_charging(self, charging_request: BotChargingRequest) -> BotChargingResponse:
        """
        Send specified bots to charging stations.
        
        This endpoint proxies the request to the brain API which handles the ROS communication.
        
        Args:
            charging_request: The request parameters including bot_idxs, force flag, 
                             block_until_done flag, and timeout
                             
        Returns:
            BotChargingResponse: Response with success/failure information
            
        Examples:
            >>> from nys_client.bot import BotChargingRequest
            >>> response = client.bots.go_charging(
            ...     BotChargingRequest(
            ...         bot_ids=["1", "2"],
            ...         force=True,
            ...         block_until_done=False,
            ...         timeout=180
            ...     )
            ... )
            >>> print(f"Success: {response.success}, Message: {response.message}")
        """
        response = self._client._request(
            method="POST",
            endpoint="api/v2/bot/go_charging",
            json=charging_request.dict()
        )
        return BotChargingResponse(**response.json())

    def stop_charging(self, stop_charging_request: BotStopChargingRequest) -> BotStopChargingResponse:
        """
        Stop charging for specified bots.
        
        This endpoint proxies the request to the brain API which handles the ROS communication.
        
        Args:
            stop_charging_request: The request parameters including bot_idxs and force flag
            
        Returns:
            BotStopChargingResponse: Response with success/failure information
            
        Examples:
            >>> from nys_client.bot import BotStopChargingRequest
            >>> response = client.bots.stop_charging(
            ...     BotStopChargingRequest(
            ...         bot_ids=["1", "2"],
            ...         force=True
            ...     )
            ... )
            >>> print(f"Success: {response.success}, Message: {response.message}")
        """
        response = self._client._request(
            method="POST",
            endpoint="api/v2/bot/stop_charging",
            json=stop_charging_request.dict()
        )
        return BotStopChargingResponse(**response.json())

    def remove_bots(self, bot_ids: List[str]) -> BotRemoveResponse:
        """
        Remove specified bots from the database.
        
        Warning: This is not the usual offboarding process. Use at your own risk!
        
        Args:
            bot_ids: List of bot IDs to remove from database; use ["*"] for all bots
            
        Returns:
            BotRemoveResponse: Response with success/failure information
            
        Examples:
            >>> response = client.bots.remove_bots(["1", "2"])
            >>> print(f"Success: {response.success}, Message: {response.message}")
        """
        response = self._client._request(
            method="DELETE",
            endpoint="api/v2/bot",
            params={"bot_ids": bot_ids}
        )
        return BotRemoveResponse(**response.json())


"""
A Python client library for the Noyes API.
            
        Example:
    >>> from nys_client import NysClient, RequestType, TriggerStatus
    >>> 
    >>> # Initialize client
    >>> client = NysClient(api_key="your_api_key")
    >>> 
    >>> # Create a request
    >>> new_request = client.requests.create( #TODO
            ...     request_type=RequestType.FULFILLMENT,
    ...     entities=[("123", 1)]
    ... )
    >>> 
    >>> # Search for jobs
    >>> jobs = client.jobs.search(
    ...     filters={"status": "EXECUTING", "type": "PICKING"}
    ... )
    >>> 
    >>> # Trigger a job
    >>> client.jobs.trigger(
    ...     job_id="job-123",
    ...     trigger_status=TriggerStatus.SUCCEEDED_TRIGGER
            ... )
        """

from .client import NysClient
from .system import SystemResource
from .bot import BotResource

# Re-export commonly used types from nys_schemas and nys_constants
from nys_constants.nys_constants import (
    RequestType,
    RequestStatus,
    JobType,
    JobStatus,
    TriggerStatus,
    SortOrder,
)

from nys_schemas.request_schema import (
    RequestResponseComposite,
    RequestFilter,
    FulfillmentRequestCreateInput,
    FetchRequestCreateInput,
    ReplenishmentRequestCreateInput,
    OnboardingRequestCreateInput,
    OffboardingRequestCreateInput,
    RFIDWriteRequestCreateInput,
    PauseRequestCreateInput,
    ResumeRequestCreateInput,
)
from nys_schemas.job_schema import (
    JobResponse,
    JobFilter,
)
from nys_schemas.bot_schema import (
    Bot,
    BotResponse,
    BotChargingRequest,
    BotChargingResponse,
    BotStopChargingRequest,
    BotStopChargingResponse,
    BotRemoveResponse,
)
from nys_schemas.load_schema import (
    LoadResponse,
    LoadCreate,
    LoadPatch,
)

__all__ = [
    # Main client class
    'NysClient',
    
    # Response models - Core entities
    'RequestResponseComposite',
    'JobResponse',
    'BotResponse',
    
    # Request/Input models
    'RequestCreateInputGeneric',
    'RequestFilter',
    'JobFilter',
    'BotChargingRequest',
    'BotChargingResponse',
    'BotStopChargingRequest', 
    'BotStopChargingResponse',
    'BotRemoveResponse',
    'Bot',
    'FetchRequestCreateInput',
    'FulfillmentRequestCreateInput',
    'ReplenishmentRequestCreateInput',
    'OnboardingRequestCreateInput',
    'OffboardingRequestCreateInput',
    'RFIDWriteRequestCreateInput',
    'PauseRequestCreateInput',
    'ResumeRequestCreateInput',
    
    # Load models
    'LoadResponse',
    'LoadCreate',
    'LoadPatch',

    
    # Enums and Constants
    'RequestType',
    'RequestStatus',
    'JobType',
    'JobStatus',
    'TriggerStatus',
    'SortOrder',
] 

__version__ = "0.1.0"